@@ -49,8 +49,7 @@
 Module for data correction and masking functions.
 """
 
-from __future__ import (absolute_import, division, print_function,
-                        unicode_literals)
+from __future__ import absolute_import, division, print_function, unicode_literals
 
 import numpy as np
 import scipy.ndimage
@@ -68,24 +67,25 @@ logger = logging.getLogger(__name__)
 __author__ = "Doga Gursoy, William Judge"
 __credits__ = "Mark Rivers, Xianghui Xiao"
 __copyright__ = "Copyright (c) 2015, UChicago Argonne, LLC."
-__docformat__ = 'restructuredtext en'
+__docformat__ = "restructuredtext en"
 __all__ = [
-    'adjust_range',
-    'circ_mask',
-    'gaussian_filter',
-    'median_filter',
-    'median_filter_cuda',
-    'median_filter_nonfinite',
-    'median_filter3d',
-    'remove_outlier3d',
-    'sobel_filter',
-    'remove_nan',
-    'remove_neg',
-    'remove_outlier',
-    'remove_outlier1d',
-    'remove_outlier_cuda',
-    'remove_ring',
-    'enhance_projs_aps_1id',
+    "adjust_range",
+    "circ_mask",
+    "gaussian_filter",
+    "median_filter",
+    "median_filter_cuda",
+    "median_filter_nonfinite",
+    "median_filter3d",
+    "remove_outlier3d",
+    "sobel_filter",
+    "remove_nan",
+    "remove_neg",
+    "remove_outlier",
+    "remove_outlier1d",
+    "remove_outlier_cuda",
+    "remove_ring",
+    "enhance_projs_aps_1id",
+    "inpainter_morph",
 ]
 
 
@@ -156,11 +156,13 @@ def gaussian_filter(arr, sigma=3, order=0, axis=0, ncore=None):
         slc = [slice(None)] * arr.ndim
         for i in range(arr.shape[axis]):
             slc[axis] = i
-            e.submit(scipy.ndimage.gaussian_filter,
-                     arr[tuple(slc)],
-                     sigma,
-                     order=order,
-                     output=out[tuple(slc)])
+            e.submit(
+                scipy.ndimage.gaussian_filter,
+                arr[tuple(slc)],
+                sigma,
+                order=order,
+                output=out[tuple(slc)],
+            )
     return out
 
 
@@ -194,10 +196,12 @@ def median_filter(arr, size=3, axis=0, ncore=None):
         slc = [slice(None)] * arr.ndim
         for i in range(arr.shape[axis]):
             slc[axis] = i
-            e.submit(scipy.ndimage.median_filter,
-                     arr[tuple(slc)],
-                     size=(size, size),
-                     output=out[tuple(slc)])
+            e.submit(
+                scipy.ndimage.median_filter,
+                arr[tuple(slc)],
+                size=(size, size),
+                output=out[tuple(slc)],
+            )
     return out
 
 
@@ -234,7 +238,7 @@ def median_filter_cuda(arr, size=3, axis=0):
 
         winAllow = range(2, 16)
 
-        if (axis != 0):
+        if axis != 0:
             arr = np.swapaxes(arr, 0, axis)
 
         if size in winAllow:
@@ -250,9 +254,11 @@ def median_filter_cuda(arr, size=3, axis=0):
             for step in range(prjsize):
                 # im_noisecu = arr[:][step][:].astype(np.float32)
                 im_noisecu = arr[step].astype(np.float32)
-                im_noisecu = np.lib.pad(im_noisecu, ((loffset, roffset),
-                                                     (loffset, roffset)),
-                                        'symmetric')
+                im_noisecu = np.lib.pad(
+                    im_noisecu,
+                    ((loffset, roffset), (loffset, roffset)),
+                    "symmetric",
+                )
                 im_noisecu = im_noisecu.flatten()
 
                 filter.setCuImage(im_noisecu)
@@ -261,7 +267,7 @@ def median_filter_cuda(arr, size=3, axis=0):
                 results = results.reshape(imsizey, imsizex)
                 out[step] = results
 
-            if (axis != 0):
+            if axis != 0:
                 out = np.swapaxes(out, 0, axis)
         else:
             warnings.warn("Window size not support, using cpu median filter")
@@ -321,7 +327,6 @@ def median_filter_nonfinite(arr, size=3, callback=None):
 
         # Iterating through each bad value and replace it with finite median
         for x_idx, y_idx in zip(*nonfinite_idx):
-
             # Determining the lower and upper bounds for kernel
             x_lower = max(0, x_idx - (size // 2))
             x_higher = min(arr.shape[1], x_idx + (size // 2) + 1)
@@ -329,21 +334,22 @@ def median_filter_nonfinite(arr, size=3, callback=None):
             y_higher = min(arr.shape[2], y_idx + (size // 2) + 1)
 
             # Extracting kernel data and fining finite median
-            kernel_cropped_arr = projection_copy[x_lower:x_higher,
-                                                 y_lower:y_higher]
+            kernel_cropped_arr = projection_copy[x_lower:x_higher, y_lower:y_higher]
 
             if len(kernel_cropped_arr[np.isfinite(kernel_cropped_arr)]) == 0:
                 raise ValueError(
                     "Found kernel containing only non-finite values.\
-                                 Please increase kernel size")
+                                 Please increase kernel size"
+                )
 
             median_corrected_arr = np.median(
-                kernel_cropped_arr[np.isfinite(kernel_cropped_arr)])
+                kernel_cropped_arr[np.isfinite(kernel_cropped_arr)]
+            )
 
             # Replacing bad data with finite median
             projection[x_idx, y_idx] = median_corrected_arr
 
-        callback(arr.shape[0], 'Nonfinite median filter', ' prjs')
+        callback(arr.shape[0], "Nonfinite median filter", " prjs")
 
     return arr
 
@@ -379,7 +385,7 @@ def median_filter3d(arr, size=3, ncore=None):
         ncore = mproc.mp.cpu_count()
 
     input_type = arr.dtype
-    if (input_type != 'float32') and (input_type != 'uint16'):
+    if (input_type != "float32") and (input_type != "uint16"):
         arr = dtype.as_float32(arr)  # silent convertion to float32 data type
     out = np.empty_like(arr)
     dif = 0.0  # set to 0 to avoid selective filtering
@@ -395,12 +401,14 @@ def median_filter3d(arr, size=3, ncore=None):
         raise ValueError("The input array must be a 3D array")
 
     # perform full 3D filtering
-    if (input_type == 'float32'):
-        extern.c_median_filt3d_float32(np.ascontiguousarray(arr), out, kernel_half_size, dif, ncore,
-                                    dx, dy, dz)
+    if input_type == "float32":
+        extern.c_median_filt3d_float32(
+            np.ascontiguousarray(arr), out, kernel_half_size, dif, ncore, dx, dy, dz
+        )
     else:
-        extern.c_median_filt3d_uint16(np.ascontiguousarray(arr), out, kernel_half_size, dif, ncore,
-                                    dx, dy, dz)
+        extern.c_median_filt3d_uint16(
+            np.ascontiguousarray(arr), out, kernel_half_size, dif, ncore, dx, dy, dz
+        )
     return out
 
 
@@ -439,9 +447,9 @@ def remove_outlier3d(arr, dif, size=3, ncore=None):
         ncore = mproc.mp.cpu_count()
 
     input_type = arr.dtype
-    if (input_type != 'float32') and (input_type != 'uint16'):
+    if (input_type != "float32") and (input_type != "uint16"):
         arr = dtype.as_float32(arr)  # silent convertion to float32 data type
-    out = np.empty_like(arr)
+    out = np.copy(arr, order="C")
 
     # convert the full kernel size (odd int) to a half size as the C function requires it
     kernel_half_size = (max(int(size), 3) - 1) // 2
@@ -454,12 +462,14 @@ def remove_outlier3d(arr, dif, size=3, ncore=None):
         raise ValueError("The input array must be a 3D array")
 
     # perform full 3D filtering
-    if (input_type == 'float32'):
-        extern.c_median_filt3d_float32(np.ascontiguousarray(arr), out, kernel_half_size, dif, ncore,
-                                    dx, dy, dz)
+    if input_type == "float32":
+        extern.c_median_filt3d_float32(
+            np.ascontiguousarray(arr), out, kernel_half_size, dif, ncore, dx, dy, dz
+        )
     else:
-        extern.c_median_filt3d_uint16(np.ascontiguousarray(arr), out, kernel_half_size, dif, ncore,
-                                    dx, dy, dz)
+        extern.c_median_filt3d_uint16(
+            np.ascontiguousarray(arr), out, kernel_half_size, dif, ncore, dx, dy, dz
+        )
     return out
 
 
@@ -495,7 +505,7 @@ def sobel_filter(arr, axis=0, ncore=None):
     return out
 
 
-def remove_nan(arr, val=0., ncore=None):
+def remove_nan(arr, val=0.0, ncore=None):
     """
     Replace NaN values in array with a given value.
 
@@ -517,12 +527,12 @@ def remove_nan(arr, val=0., ncore=None):
     val = np.float32(val)
 
     with mproc.set_numexpr_threads(ncore):
-        ne.evaluate('where(arr!=arr, val, arr)', out=arr)
+        ne.evaluate("where(arr!=arr, val, arr)", out=arr)
 
     return arr
 
 
-def remove_neg(arr, val=0., ncore=None):
+def remove_neg(arr, val=0.0, ncore=None):
     """
     Replace negative values in array with a given value.
 
@@ -544,7 +554,7 @@ def remove_neg(arr, val=0., ncore=None):
     val = np.float32(val)
 
     with mproc.set_numexpr_threads(ncore):
-        ne.evaluate('where(arr<0, val, arr)', out=arr)
+        ne.evaluate("where(arr<0, val, arr)", out=arr)
     return arr
 
 
@@ -587,17 +597,19 @@ def remove_outlier(arr, dif, size=3, axis=0, ncore=None, out=None):
         slc = [slice(None)] * arr.ndim
         for i in range(ncore):
             slc[axis] = chnk_slices[i]
-            e.submit(scipy.ndimage.median_filter,
-                     arr[tuple(slc)],
-                     size=filt_size,
-                     output=tmp[tuple(slc)])
+            e.submit(
+                scipy.ndimage.median_filter,
+                arr[tuple(slc)],
+                size=filt_size,
+                output=tmp[tuple(slc)],
+            )
 
     arr = dtype.as_float32(arr)
     tmp = dtype.as_float32(tmp)
     dif = np.float32(dif)
 
     with mproc.set_numexpr_threads(ncore):
-        out = ne.evaluate('where(arr-tmp>=dif,tmp,arr)', out=out)
+        out = ne.evaluate("where(arr-tmp>=dif,tmp,arr)", out=out)
 
     return out
 
@@ -637,8 +649,7 @@ def remove_outlier1d(arr, dif, size=3, axis=0, ncore=None, out=None):
     other_axes = [i for i in range(arr.ndim) if i != axis]
     largest = np.argmax([arr.shape[i] for i in other_axes])
     lar_axis = other_axes[largest]
-    ncore, chnk_slices = mproc.get_ncore_slices(arr.shape[lar_axis],
-                                                ncore=ncore)
+    ncore, chnk_slices = mproc.get_ncore_slices(arr.shape[lar_axis], ncore=ncore)
     filt_size = [1] * arr.ndim
     filt_size[axis] = size
 
@@ -646,14 +657,16 @@ def remove_outlier1d(arr, dif, size=3, axis=0, ncore=None, out=None):
         slc = [slice(None)] * arr.ndim
         for i in range(ncore):
             slc[lar_axis] = chnk_slices[i]
-            e.submit(scipy.ndimage.median_filter,
-                     arr[slc],
-                     size=filt_size,
-                     output=tmp[slc],
-                     mode='mirror')
+            e.submit(
+                scipy.ndimage.median_filter,
+                arr[slc],
+                size=filt_size,
+                output=tmp[slc],
+                mode="mirror",
+            )
 
     with mproc.set_numexpr_threads(ncore):
-        out = ne.evaluate('where(arr-tmp>=dif,tmp,arr)', out=out)
+        out = ne.evaluate("where(arr-tmp>=dif,tmp,arr)", out=out)
 
     return out
 
@@ -698,7 +711,7 @@ def remove_outlier_cuda(arr, dif, size=3, axis=0):
 
         winAllow = range(2, 16)
 
-        if (axis != 0):
+        if axis != 0:
             arr = np.swapaxes(arr, 0, axis)
 
         if size in winAllow:
@@ -713,9 +726,9 @@ def remove_outlier_cuda(arr, dif, size=3, axis=0):
 
             for step in range(prjsize):
                 im_noisecu = arr[step].astype(np.float32)
-                im_noisecu = np.lib.pad(im_noisecu, ((loffset, roffset),
-                                                     (loffset, roffset)),
-                                        'symmetric')
+                im_noisecu = np.lib.pad(
+                    im_noisecu, ((loffset, roffset), (loffset, roffset)), "symmetric"
+                )
                 im_noisecu = im_noisecu.flatten()
 
                 filter.setCuImage(im_noisecu)
@@ -724,7 +737,7 @@ def remove_outlier_cuda(arr, dif, size=3, axis=0):
                 results = results.reshape(imsizey, imsizex)
                 out[step] = results
 
-            if (axis != 0):
+            if axis != 0:
                 out = np.swapaxes(out, 0, axis)
         else:
             warnings.warn("Window size not support, using cpu outlier removal")
@@ -737,18 +750,20 @@ def remove_outlier_cuda(arr, dif, size=3, axis=0):
     return out
 
 
-def remove_ring(rec,
-                center_x=None,
-                center_y=None,
-                thresh=300.0,
-                thresh_max=300.0,
-                thresh_min=-100.0,
-                theta_min=30,
-                rwidth=30,
-                int_mode='WRAP',
-                ncore=None,
-                nchunk=None,
-                out=None):
+def remove_ring(
+    rec,
+    center_x=None,
+    center_y=None,
+    thresh=300.0,
+    thresh_max=300.0,
+    thresh_min=-100.0,
+    theta_min=30,
+    rwidth=30,
+    int_mode="WRAP",
+    ncore=None,
+    nchunk=None,
+    out=None,
+):
     """
     Remove ring artifacts from images in the reconstructed domain.
     Descriptions of parameters need to be more clear for sure.
@@ -803,9 +818,9 @@ def remove_ring(rec,
     if center_y is None:
         center_y = (dy - 1.0) / 2.0
 
-    if int_mode.lower() == 'wrap':
+    if int_mode.lower() == "wrap":
         int_mode = 0
-    elif int_mode.lower() == 'reflect':
+    elif int_mode.lower() == "reflect":
         int_mode = 1
     else:
         raise ValueError("int_mode should be WRAP or REFLECT")
@@ -813,19 +828,30 @@ def remove_ring(rec,
     if not 0 <= theta_min < 180:
         raise ValueError("theta_min should be in the range [0 - 180)")
 
-    args = (center_x, center_y, dx, dy, dz, thresh_max, thresh_min, thresh,
-            theta_min, rwidth, int_mode)
+    args = (
+        center_x,
+        center_y,
+        dx,
+        dy,
+        dz,
+        thresh_max,
+        thresh_min,
+        thresh,
+        theta_min,
+        rwidth,
+        int_mode,
+    )
 
     axis_size = rec.shape[0]
     ncore, nchunk = mproc.get_ncore_nchunk(axis_size, ncore, nchunk)
     with cf.ThreadPoolExecutor(ncore) as e:
         for offset in range(0, axis_size, nchunk):
-            slc = np.s_[offset:offset + nchunk]
+            slc = np.s_[offset : offset + nchunk]
             e.submit(extern.c_remove_ring, out[slc], *args)
     return out
 
 
-def circ_mask(arr, axis, ratio=1, val=0., ncore=None):
+def circ_mask(arr, axis, ratio=1, val=0.0, ncore=None):
     """
     Apply circular mask to a 3D array.
 
@@ -853,7 +879,7 @@ def circ_mask(arr, axis, ratio=1, val=0., ncore=None):
     mask = _get_mask(dy, dz, ratio)
 
     with mproc.set_numexpr_threads(ncore):
-        ne.evaluate('where(mask, _arr, val)', out=_arr)
+        ne.evaluate("where(mask, _arr, val)", out=_arr)
 
     return _arr.swapaxes(0, axis)
 
@@ -876,13 +902,13 @@ def _get_mask(dx, dy, ratio):
     ndarray
         2D boolean array.
     """
-    rad1 = dx / 2.
-    rad2 = dy / 2.
+    rad1 = dx / 2.0
+    rad2 = dy / 2.0
     if dx < dy:
         r2 = rad1 * rad1
     else:
         r2 = rad2 * rad2
-    y, x = np.ogrid[0.5 - rad1:0.5 + rad1, 0.5 - rad2:0.5 + rad2]
+    y, x = np.ogrid[0.5 - rad1 : 0.5 + rad1, 0.5 - rad2 : 0.5 + rad2]
     return x * x + y * y < ratio * ratio * r2
 
 
@@ -914,11 +940,13 @@ def enhance_projs_aps_1id(imgstack, median_ks=5, ncore=None):
     tmp = []
     with cf.ProcessPoolExecutor(ncore) as e:
         for n_img in range(imgstack.shape[0]):
-            tmp.append(e.submit(
-                _enhance_img,
-                imgstack[n_img, :, :],
-                median_ks,
-            ))
+            tmp.append(
+                e.submit(
+                    _enhance_img,
+                    imgstack[n_img, :, :],
+                    median_ks,
+                )
+            )
 
     return np.stack([me.result() for me in tmp], axis=0)
 
@@ -945,7 +973,7 @@ def _enhance_img(img, median_ks, normalized=True):
     """
     wgt = _calc_histequal_wgt(img)
     img = medfilt2d(img, kernel_size=median_ks).astype(np.float64)
-    img = ne.evaluate('(img**2)*wgt', out=img)
+    img = ne.evaluate("(img**2)*wgt", out=img)
     return img / img.max() if normalized else img
 
 
@@ -965,3 +993,166 @@ def _calc_histequal_wgt(img):
         image
     """
     return (np.sort(img.flatten()).searchsorted(img) + 1) / np.prod(img.shape)
+
+
+def inpainter_morph(
+    arr,
+    mask,
+    size=5,
+    iterations=2,
+    inpainting_type="random",
+    axis=None,
+    ncore=None,
+):
+    """
+    Apply 3D or 2D morphological inpainter (extrapolator) to a given missing
+    data array using provided mask (boolean). The algorithm is presented in the
+    paper :cite:`Kazantsev:23`.
+
+    If applied to 3D tomographic data, one can use the 2D module applied to a
+    sinogram by specifing the axis in the list of parameters bellow.
+    Alternatively, one can use the 3D module with symmetric 3D kernels
+    (axis=None). The latter is more recommended for 3D data as it ensures the
+    smoother intensity transition in every direction.
+
+
+    .. versionadded:: 1.15
+
+    Parameters
+    ----------
+    arr : ndarray
+        Input 3D or 2D array of float32 data type with the missing data.
+    mask : ndarray, bool
+        Boolean mask array of the same size as arr. True values in the mask
+        indicate the region that needs to be inpainted.
+    size : int, optional
+        The size of the searching window (a kernel).
+    iterations : int, optional
+        Iterations of the algorithm after the inpainted region was processed
+        fully. The larger numbers usually lead to oversmoothing of the area, we
+        recommend 2.
+    inpainting_type : str, optional
+        The type of the inpainting technique. Choose between 'mean', 'median'
+        or 'random' neighbour selection. We suggest using 'random' to minimise
+        the "spilling" of intensities inside the inpainted areas.
+    axis : int, optional
+        Split the data along this axis, and perform inpainting in the remaining
+        dimensions. If set to `None` then the 3D inpainting is enabled, which
+        is recommended to use for 3D data. If the 2D inpainting used for 3D
+        data, it is best to apply it in sinogram space.
+    ncore : int, optional
+        Number of cores that will be assigned to jobs. All cores will be used
+        if unspecified.
+
+    Returns
+    -------
+    ndarray
+       Inpainted array of float32 data type.
+
+    Raises
+    ------
+    ValueError
+        If the input array is not float32.
+
+        If the input mask array is not bool.
+
+        If inpainting_type is not 'mean', 'median' or 'random'.
+
+        If axis is an integer for 2D input data.
+
+        If axis is outside of the accepted range.
+
+    """
+    if ncore is None:
+        ncore = mproc.mp.cpu_count()
+
+    if arr.dtype != "float32":
+        arr = dtype.as_float32(arr)  # silent convertion to float32 data type
+    if mask.dtype != "bool":
+        raise ValueError(
+            "Please make sure that the mask provided is boolean where true values define the missing data regions!"
+        )
+    out = np.empty_like(arr, order="C")
+
+    if inpainting_type not in ["mean", "median", "random"]:
+        raise ValueError(
+            "Inpainting type should be chosen as 'mean', 'median' or 'random'"
+        )
+
+    if inpainting_type == "mean":
+        inp_type_int = 0
+    if inpainting_type == "median":
+        inp_type_int = 1
+    if inpainting_type == "random":
+        inp_type_int = 2
+
+    if axis is not None:
+        if arr.ndim == 2:
+            raise ValueError("The axis number is valid only for 3D data, please set axis to None for 2D data")
+        else:
+            if axis < 0 or axis > 2:
+                raise ValueError("The accepted axes range for 3D data is [0:2]")
+
+    if arr.ndim == 3:
+        dz, dy, dx = arr.shape
+        if (dz == 0) or (dy == 0) or (dx == 0):
+            raise ValueError("The length of one of dimensions is equal to zero")
+        if axis is not None:
+            # perform 2d inpainting for 3d data using the provided axis
+            slices_list = [slice(None), slice(None), slice(None)]
+            for m in range(arr.shape[axis]):
+                slices_list[axis] = slice(m,m+1)
+                slice2d = arr[tuple(slices_list)].squeeze()
+                mask2d = mask[tuple(slices_list)].squeeze()
+                dy, dx = slice2d.shape
+                out2d = np.empty_like(slice2d, order="C")
+                extern.c_inpainter(
+                    np.ascontiguousarray(slice2d),
+                    np.ascontiguousarray(mask2d),
+                    out2d,
+                    iterations,
+                    size,
+                    inp_type_int,
+                    ncore,
+                    dx,
+                    dy,
+                    1,
+                )
+                if axis == 0:
+                    out[m, :, :] = out2d
+                elif axis == 1:
+                    out[:, m, :] = out2d
+                else:
+                    out[:, :, m] = out2d
+        else:
+            # 3d inpaiting for 3d data
+            extern.c_inpainter(
+                np.ascontiguousarray(arr),
+                np.ascontiguousarray(mask),
+                out,
+                iterations,
+                size,
+                inp_type_int,
+                ncore,
+                dx,
+                dy,
+                dz,
+            )
+    else:
+        dy, dx = arr.shape
+        if (dy == 0) or (dx == 0):
+            raise ValueError("The length of one of dimensions is equal to zero")
+        extern.c_inpainter(
+            np.ascontiguousarray(arr),
+            np.ascontiguousarray(mask),
+            out,
+            iterations,
+            size,
+            inp_type_int,
+            ncore,
+            dx,
+            dy,
+            1,
+        )
+
+    return out
