@@ -502,7 +502,7 @@ def fan_to_para(tomo, dist, geom):
     ndarray
         Transformed 3D tomographic data.
     """
-    logger.warning('Not implemented.')
+    raise NotImplementedError('Not implemented.')
 
 
 def para_to_fan(tomo, dist, geom):
@@ -529,7 +529,7 @@ def para_to_fan(tomo, dist, geom):
     ndarray
         Transformed 3D tomographic data.
     """
-    logger.warning('Not implemented.')
+    raise NotImplementedError('Not implemented.')
 
 
 def add_focal_spot_blur(tomo, spotsize):
@@ -547,7 +547,7 @@ def add_focal_spot_blur(tomo, spotsize):
     spotsize : float
         Focal spot size of circular x-ray source.
     """
-    logger.warning('Not implemented.')
+    raise NotImplementedError('Not implemented.')
 
 
 def _get_magnification(r1, r2):
@@ -592,4 +592,4 @@ def _get_otf(dx, dy, px, py, spotsize):
     array
         2D OTF function.
     """
-    logger.warning('Not implemented.')
+    raise NotImplementedError('Not implemented.')
